/*
 * Licensed Materials - Property of IBM
 * 5648-C63
 * (C) Copyright IBM Corp. 1999, 2000
 */

/******************************************************************************/
/*                                                                            */
/* Note: This sample plugin node implementation is not necessarily portable   */
/* to all platforms supported by MQSeries Integrator Version 2.               */
/*                                                                            */
/******************************************************************************/

/* Includes */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>

#include <cmqc.h>
#include <PublishExceptionNode.h>

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  Acquire storage for attribute and link to the context                     */
/*                                                                            */
/******************************************************************************/
ATTRIBUTE_TBL_ENTRY *insAttrTblEntry(
  NODE_CONTEXT_ST* context,
  CciChar*         attrName,
  int              type
){
  static char* functionName = "insAttrTblEntry()";
  ATTRIBUTE_TBL_ENTRY* entry;
  int                  attrSize;

  entry = malloc(sizeof(ATTRIBUTE_TBL_ENTRY));
  if (entry) {
    entry->next = 0;
    entry->type = type;

    switch (entry->type) {
    case CNI_TYPE_INTEGER:
      attrSize = 4;
      break;
    case CNI_TYPE_STRING:
      attrSize = MAX_STRING_SIZE;
    default:
      /* Error: Invalid datatype */
      break;
    }

    /* Acquire and clear storage for attribute value */
    entry->valptr = malloc(attrSize);
    if (entry->valptr) memset(entry->valptr, 0, attrSize);

    /* Store the attribute name */
#if defined(__CCI_SOLARIS__)
    CciCharCpy(entry->name, attrName);
#else
    wcscpy(entry->name, attrName);
#endif

    /* Link an existing previous element to this one */
    if (context->attrListPrevious) context->attrListPrevious->next = entry;
    else if ((context->attrListHead) == 0) context->attrListHead = entry;

    /* Save the pointer to the previous element */
    context->attrListPrevious = entry;
  }
  else {
    /* Error: Unable to allocate memory */
  }

  return(entry);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Return the terminal object for a named input terminal                     */
/*                                                                            */
/******************************************************************************/
void *getInputTerminalHandle(
  NODE_CONTEXT_ST* context,
  CciChar*         terminalName
){
  static char* functionName = "getInputTerminalHandle()";
  TERMINAL_LIST_ENTRY *p = ((NODE_CONTEXT_ST *)context)->inputTerminalListHead;

  /* Scan the list for a matching terminal name */
  while (p != 0) {
#if defined(__CCI_SOLARIS__)
    if (CciCharCmp(p->name, terminalName) == 0) {
#else
    if (wcscmp(p->name, terminalName) == 0) {
#endif
      /* Match found; return it */
      return(p->handle);
    }
    p = p->next;
  }

  /* No match was found */
  return(0);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Acquire storage for an input terminal entry and link it to the context    */
/*                                                                            */
/******************************************************************************/
TERMINAL_LIST_ENTRY *insInputTerminalListEntry(
  NODE_CONTEXT_ST* context,
  CciChar*         terminalName 
){
  static char* functionName = "insInputTerminalListEntry()";
  TERMINAL_LIST_ENTRY* entry;
  int                  rc;

  entry = malloc(sizeof(TERMINAL_LIST_ENTRY));
  if (entry) {

    /* This entry is the current end of the list */
    entry->next = 0;

    /* Store the terminal name */
#if defined(__CCI_SOLARIS__)
    CciCharCpy(entry->name, terminalName);
#else
    wcscpy(entry->name, terminalName);
#endif

    /* Create terminal and save its handle */
    entry->handle = cniCreateInputTerminal(&rc, context->nodeObject, (CciChar*)terminalName);

    /* Link an existing previous element to this one */
    if (context->inputTerminalListPrevious) context->inputTerminalListPrevious->next = entry;
    else if ((context->inputTerminalListHead) == 0) context->inputTerminalListHead = entry;

    /* Save the pointer to the previous element */
    context->inputTerminalListPrevious = entry;
  }
  else {
    /* Error: Unable to allocate memory */
  }

  return(entry);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Return the terminal object for a named output terminal                    */
/*                                                                            */
/******************************************************************************/
void *getOutputTerminalHandle(
  NODE_CONTEXT_ST* context,
  CciChar*         terminalName
){
  static char* functionName = "getOutputTerminalHandle()";
  TERMINAL_LIST_ENTRY *p = ((NODE_CONTEXT_ST *)context)->outputTerminalListHead;

  /* Scan the list for a matching terminal name */
  while (p != 0) {
#if defined(__CCI_SOLARIS__)
    if (CciCharCmp(p->name, terminalName) == 0) {
#else
    if (wcscmp(p->name, terminalName) == 0) {
#endif
      /* Match found; return it */
      return(p->handle);
    }
    p = p->next;
  }

  /* No match was found */
  return(0);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Acquire storage for an output terminal and link it to the context         */
/*                                                                            */
/******************************************************************************/
TERMINAL_LIST_ENTRY *insOutputTerminalListEntry(
  NODE_CONTEXT_ST* context,
  CciChar*         terminalName 
){
  static char* functionName = "insOutputTerminalListEntry()";
  TERMINAL_LIST_ENTRY* entry;
  int                  rc;

  entry = malloc(sizeof(TERMINAL_LIST_ENTRY));
  if (entry) {

    /* This entry is the current end of the list */
    entry->next = 0;

    /* Store the terminal name */
#if defined(__CCI_SOLARIS__)
    CciCharCpy(entry->name, terminalName);
#else
    wcscpy(entry->name, terminalName);
#endif

    /* Create terminal and save its handle */
    entry->handle = cniCreateOutputTerminal(&rc, context->nodeObject, (CciChar*)terminalName);

    /* Link an existing previous element to this one */
    if (context->outputTerminalListPrevious) context->outputTerminalListPrevious->next = entry;
    else if ((context->outputTerminalListHead) == 0) context->outputTerminalListHead = entry;

    /* Save the pointer to the previous element */
    context->outputTerminalListPrevious = entry;
  }
  else {
    /* Error: Unable to allocate memory */
  }

  return(entry);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                cniCreateNodeContext() */
/*                                                                            */
/******************************************************************************/
CciContext* _PublishException_createNodeContext(
  CciFactory* factoryObject,
  CciChar*    nodeName,
  CciNode*    nodeObject
){
  static char* functionName = "_PublishException_createNodeContext()";
  NODE_CONTEXT_ST* p;
  CciChar          buffer[256];
  char             filename[256];

  /* Allocate a pointer to the local context */
  p = malloc(sizeof(NODE_CONTEXT_ST));

  if (p) {

    /* Clear the context area */
    memset(p, 0, sizeof(NODE_CONTEXT_ST));

    /* Save our node object pointer in our context */
    p->nodeObject = nodeObject;

    /* Save our node name */
#if defined(__CCI_SOLARIS__)
    CciCharNCpy((CciChar*)&p->nodeName, nodeName, MAX_NODE_NAME_LEN);
#else
    wcsncpy((wchar_t*)&p->nodeName, nodeName, MAX_NODE_NAME_LEN);
#endif

    /* Create attributes and set default values */   
    /* Attribute value set at start time. */
    insAttrTblEntry(p, (CciChar *)constUserIdString, CNI_TYPE_STRING);
    
    /* Create terminals */
    insInputTerminalListEntry(p, (CciChar *)constInTerminal);
    insOutputTerminalListEntry(p, (CciChar *)constOutTerminal);
    insOutputTerminalListEntry(p, (CciChar *)constPublishTerminal);
    insOutputTerminalListEntry(p, (CciChar *)constFailureTerminal);

  }
  else {
    /* Error: Unable to allocate memory */
  }

  /* Return the pointer to the local context */
  return((CciContext*)p);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                cniDeleteNodeContext() */
/*                                                                            */
/* This function is called when an instance of a node is deleted.             */
/*                                                                            */
/******************************************************************************/
void _deleteNodeContext(
  CciContext* context
){
  static char* functionName = "_deleteNodeContext()";

  return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniGetAttribute() */
/*                                                                            */
/******************************************************************************/
int _getAttributeName(
  CciContext* context,
  int         index,
  CciChar*    buffer,
  int         bufsize
){
  static char* functionName = "_getAttributeName()";
  int                  count = 0;
  ATTRIBUTE_TBL_ENTRY *p = ((NODE_CONTEXT_ST *)context)->attrListHead;
    
  while (p != 0) {
    if (count == index) {
#if defined(__CCI_SOLARIS__)
      CciCharCpy(buffer, p->name);
#else
      wcscpy(buffer, p->name);
#endif
      /* Normal return; attribute name returned */
      return(0);
    }
    count++;
    p = p->next;
  }

  /* Attribute not owned by the plugin node; defer to base */
  return(1);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniSetAttribute() */
/*                                                                            */
/******************************************************************************/
int _setAttribute(
  CciContext* context,
  CciChar*    attrName,
  CciChar*    attrValue
){
  static char* functionName = "_setAttribute()";
  ATTRIBUTE_TBL_ENTRY* p = ((NODE_CONTEXT_ST *)context)->attrListHead;
  NODE_CONTEXT_ST*     nc = context;        

  while (p != 0) {
#if defined(__CCI_SOLARIS__)
    if (CciCharCmp(p->name, attrName) == 0) {
#else
    if (wcscmp(p->name, attrName) == 0) {
#endif
      switch(p->type) {
      case CNI_TYPE_INTEGER:
#if defined(__CCI_SOLARIS__)
        {
          const char* mbAttrValue = mbString(attrValue);
          *((int *)p->valptr) = atoi(mbAttrValue);
          free((void *)mbAttrValue);
        }
#elif defined(__CCI_AIX__)
        *((int *)p->valptr) = watoi(attrValue);
#else
        *((int *)p->valptr) = _wtoi(attrValue);
#endif
        break;
      case CNI_TYPE_STRING:
#if defined(__CCI_SOLARIS__)
        {
          const char* mbAttrValue = mbString(attrValue);
          memcpy(p->valptr, mbAttrValue, MAX_STRING_SIZE);
          free((void *)mbAttrValue);
        }
#else
        memcpy(p->valptr, attrValue, MAX_STRING_SIZE);
#endif
      default:
        /* Error: Invalid datatype */
        break;
      }

      /* Normal return; attribute value stored by plugin */
      return(0);
    }
    p = p->next;
  }

  /* Attribute not owned by plugin */
  return(1);
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                     cniGetAttribute() */
/*                                                                            */
/******************************************************************************/
int _getAttribute(
  CciContext* context,
  CciChar*    attrName,
  CciChar*    buffer,
  int         bufsize      
){
  static char* functionName = "_getAttribute()";
  ATTRIBUTE_TBL_ENTRY* p = ((NODE_CONTEXT_ST *)context)->attrListHead;
  NODE_CONTEXT_ST*     nc = context;
  wchar_t              wstring[256];
  char                 string[256];  

  while (p != 0) {
#if defined(__CCI_SOLARIS__)
    if (CciCharCmp(p->name, attrName) == 0) {
#else
    if (wcscmp(p->name, attrName) == 0) {
#endif
      switch(p->type) {
      case CNI_TYPE_INTEGER:
        sprintf((char*)&string, "%d", *((int*)p->valptr));
#if defined(__CCI_SOLARIS__)
        {
          const CciChar* ucsAttrValue = CciString(string);
          CciCharNCpy(buffer, ucsAttrValue, min(bufsize, 256));
          free((void *)ucsAttrValue);
        }
#else
        mbstowcs((wchar_t*)&wstring, (char*)&string, 256); 
        wcsncpy(buffer, (wchar_t*)&wstring, min(bufsize, 256));
#endif
        break;
      case CNI_TYPE_STRING:
#if defined(__CCI_SOLARIS__)
        {
          const CciChar* ucsAttrValue = CciString(p->valptr);
          CciCharNCpy(buffer, ucsAttrValue, min(MAX_STRING_SIZE, bufsize));
          free((void *)ucsAttrValue);
        }
#else
        memcpy(buffer, p->valptr, min(MAX_STRING_SIZE, bufsize));
#endif
      default:
        /* Error: Invalid datatype */
        break;
      }

      /* Normal return; attribute value returned */ 
      return(0);
    }
    p = p->next;
  }

  /* Attribute not owned by plugin */
  return(1); 
}

#if defined(__CCI_SOLARIS__)
/******************************************************************************/
/*                                                                            */
/* Method: CciCharCmp                                                         */
/*                                                                            */
/******************************************************************************/
int CciCharCmp(const CciChar * s1, const CciChar* s2)
{
  CciChar c1 ;
  CciChar c2 ;
  for (;;)
    {
      c1 = *s1 ;
      s1 += 1 ;
      c2 = *s2 ;
      s2 += 1 ;
      if ( (0 == c1) || (c1 != c2) ) break ;
    }
  return c1-c2 ;
}

/******************************************************************************/
/*                                                                            */
/* Method: mbString                                                          */
/*                                                                            */
/******************************************************************************/
const char* mbString(const CciChar* source)
{
  int charCount = CciCharLen(source) ;
  int maxBytes = (charCount+1) * 9 ;
  char * buffer = (char *)malloc(maxBytes) ;
  int rc ;
  cciUcsToMbs(&rc, source, buffer, maxBytes, 1208) ;
  return buffer ;
}

/******************************************************************************/
/*                                                                            */
/* Method: CciCharNCpy                                                        */
/*                                                                            */
/******************************************************************************/
void CciCharNCpy(CciChar* target, const CciChar* source, unsigned int size)
{
  CciChar c ;
  int i;
  for (i=0;i<size;i+=1)
    {
      c = *source ;
      *target = c ;
      target += 1 ;
      if ( 0 != c ) source += 1 ;
    }
}


#endif

/******************************************************************************/
/*                                                                            */
/* Method: CciString                                                          */
/*                                                                            */
/******************************************************************************/
CciChar* CciString(const char* source)
{
  /* Maximum number of characters in Unicode representation */
  int maxChars = strlen(source) + 1 ;
  CciChar* buffer = (CciChar*)malloc(maxChars * sizeof(CciChar)) ;
  int rc ;
  /* We don't know the compiler's code page at this stage, but 1208 (utf8 Unicode) */
  /* will generally be appropriate for ASCII systems                               */
  cciMbsToUcs(&rc, source, buffer, maxChars, 1208) ;
  return buffer ;
}


/******************************************************************************/
/*                                                                            */
/* Method: createIntElement                                                   */
/*                                                                            */
/******************************************************************************/
void createIntElement(
  CciElement *attachElement,
  CciChar    *elementName,
  CciInt      intValue)
{
  static char* functionName = "createIntElement()";

  CciElement*      elementChild;
  int rc = 0;
          
  /* Create the element as the last child */
  elementChild = cniCreateElementAsLastChild(&rc,
                                             attachElement);
  CCI_CHECK_RC();

  /* Set the element name */
  cniSetElementName(&rc,
                    elementChild,
                    elementName);  
  CCI_CHECK_RC();

  /* Set the integer value */
  cniSetElementIntegerValue(&rc,
                            elementChild,
                            intValue);
  CCI_CHECK_RC();

  return;
}

/******************************************************************************/
/*                                                                            */
/* Method: createBoolElement                                                  */
/*                                                                            */
/******************************************************************************/
void createBoolElement(
  CciElement *attachElement,
  CciChar    *elementName,
  CciBool     boolValue)
{
  static char* functionName = "createBoolElement()";

  CciElement*      elementChild;
  int rc = 0;
          
  /* Create the element as the last child */
  elementChild = cniCreateElementAsLastChild(&rc,
                                             attachElement);
  CCI_CHECK_RC();

  /* Set the element name */
  cniSetElementName(&rc,
                    elementChild,
                    elementName);  
  CCI_CHECK_RC();

  /* Set the integer value */
  cniSetElementBooleanValue(&rc,
                            elementChild,
                            boolValue);
  CCI_CHECK_RC();

  return;
}

/******************************************************************************/
/*                                                                            */
/* Method: createCharElement                                                  */
/*                                                                            */
/******************************************************************************/
void createCharElement(
      CciElement *attachElement,
      CciChar    *elementName,
const CciChar    *charValue,
      CciSize     charSize)
{
  static char* functionName = "createCharElement()";

  CciElement*      elementChild;
  int rc = 0;
          
  /* Create the element as the last child */
  elementChild = cniCreateElementAsLastChild(&rc,
                                             attachElement);
  CCI_CHECK_RC();

  /* Set the element name */
  cniSetElementName(&rc,
                    elementChild,
                    elementName);  
  CCI_CHECK_RC();

  /* Set the integer value */
  cniSetElementCharacterValue(&rc,
                              elementChild,
                              charValue,
                              charSize);
  CCI_CHECK_RC();

  return;
}

/******************************************************************************/
/*                                                                            */
/* Method: CciCharCpy                                                         */
/*                                                                            */
/* Copies one string into another.                                            */
/*                                                                            */
/******************************************************************************/
void CciCharCpy(CciChar* target, const CciChar* source)
{
  CciChar c ;
  for (;;)
    {
      c = *source ;
      source += 1 ;
      *target = c ;
      target += 1 ;
      if ( 0 == c ) break ;
    }
}

/******************************************************************************/
/*                                                                            */
/* Method: CciCharLen                                                         */
/*                                                                            */
/******************************************************************************/
int CciCharLen(const CciChar* s)
{
  int i=0 ;
  while (*s != 0 )
    {
      s += 1 ;
      i += 1 ;
    }
  return i ;
}

/******************************************************************************/
/*                                                                            */
/* Method: CciCharCat                                                         */
/*                                                                            */
/* Appends the source string to the target string                             */
/*                                                                            */
/******************************************************************************/
void CciCharCat(CciChar* target, const CciChar* source)
{
  CciChar c ;
  int targetLen = CciCharLen(target);

  target += targetLen;

  for (;;)
    {
      c = *source ;
      source += 1 ;
      *target = c ;
      target += 1 ;
      if ( 0 == c ) break ;
    }
}


/******************************************************************************/
/*                                                                            */
/* Method: _PublishException_evaluate()                                       */
/*                                                                            */
/* The entry point for processing a message                                   */
/*                                                                            */
/******************************************************************************/
void _PublishException_evaluate(
  CciContext* context, 
  CciMessage* destinationList,
  CciMessage* exceptionList,
  CciMessage* message
){
  static char* functionName = "_PublishException_evaluate()";

  CciTerminal*     terminalObject;
  CciMessage*      publishMsg = NULL;
  
  CciElement*      outRootElement;
  CciElement*      propertiesChild;
  CciElement*      mqmdChild;
  CciElement*      rfh2Child;
  CciElement*      pscChild;
  CciElement*      xmlChild;
  CciElement*      brokerChild;
  CciElement*      executionGroupChild;
  CciElement*      messageflowChild;
  CciElement*      exceptionChild;
  CciElement*      commandChild;
  CciElement*      commandValueChild;
  CciElement*      topicChild;
  CciElement*      topicValueChild;

  CciChar*         topic;
  CciChar*         fileName;
  CciChar*         exceptionFunctionName;
  CciChar*         traceText;
  CciChar          buffer[256]; 

  int              rc = 0; 
  int              retvalue;
  int              retvaluePublish;
  int              numInserts = 0;
  int              sizeOfTopicString = 0;
  CCI_EXCEPTION_ST exception_st; 
  CNI_BROKER_INFO_ST broker_info_st;

  /* Propagate the message to the output terminal */
  /* Get handle of output terminal */
  terminalObject = getOutputTerminalHandle(context, (CciChar*)constOutTerminal);

  /* If the terminal exists and is attached, propagate to it */
  if (terminalObject) 
  {
    if (cniIsTerminalAttached(&rc, terminalObject)) 
    {
      /* Propagate the original message to the output terminal */
      retvalue = cniPropagate(&rc, terminalObject, destinationList, exceptionList, message);
      if (retvalue == CCI_FAILURE) 
      {
        if (rc == CCI_EXCEPTION) 
        {
          /* Get details of the exception */
          memset(&exception_st, 0, sizeof(exception_st));
          cciGetLastExceptionData(&rc, &exception_st);

          /* Setup the BROKER_INFO_ST to query our environment */
          memset(&broker_info_st, 0, sizeof(broker_info_st));
          broker_info_st.brokerName.buffer               = malloc(MAX_STRING_SIZE);
          broker_info_st.brokerName.bufferLength         = MAX_STRING_SIZE;
          broker_info_st.executionGroupName.buffer       = malloc(MAX_STRING_SIZE);
          broker_info_st.executionGroupName.bufferLength = MAX_STRING_SIZE;
          broker_info_st.messageFlowName.buffer          = malloc(MAX_STRING_SIZE);
          broker_info_st.messageFlowName.bufferLength    = MAX_STRING_SIZE;
          broker_info_st.queueManagerName.buffer         = malloc(MAX_STRING_SIZE);
          broker_info_st.queueManagerName.bufferLength   = MAX_STRING_SIZE;

          /* Get details about the broker environment */
          cniGetBrokerInfo(0, ((NODE_CONTEXT_ST *)context)->nodeObject, &broker_info_st);

          /* Get details of the exception */
          memset(&exception_st, 0, sizeof(exception_st));
          cciGetLastExceptionData(&rc, &exception_st);

          /* Create a publication message to put the exception */
          /* data into                                         */
          /* Create a new message for output */
          publishMsg = cniCreateMessage(&rc, cniGetMessageContext(&rc, message));
          CCI_CHECK_RC();

          /* Get the root element of the output message */
          outRootElement = cniRootElement(&rc, publishMsg);
          CCI_CHECK_RC();

          /* Create a Properties parser */
          propertiesChild = cniCreateElementAsFirstChildUsingParser(&rc,
                                                                    outRootElement,
                                                                    (CciChar*)constMqPropertyParser);
          CCI_CHECK_RC();

          /* Create a MQMD parser */
          mqmdChild = cniCreateElementAsLastChildUsingParser(&rc, 
                                                             outRootElement,
                                                             (CciChar*)constMQHMDParser);
          CCI_CHECK_RC();

          /* From root, create an RFH2 parser */
          rfh2Child = cniCreateElementAsLastChildUsingParser(&rc,
                                                             outRootElement,
                                                             (CciChar*)constMQHRF2Parser);
          CCI_CHECK_RC();

          /* Create an XML body under root */
          xmlChild = cniCreateElementAsLastChildUsingParser(&rc,
                                                            outRootElement,
                                                            (CciChar*)constXMLParser);

          /* Actions under MQMD parser */
          
          /* Create Format name-value - RFH2 format */          
          createCharElement(mqmdChild, (CciChar*)constFormat, (CciChar*)constMQHRF2Parser, constFormatLength);

          /* MSG Type name-value */
          createIntElement(mqmdChild, (CciChar*)constMsgType, (CciInt)MQMT_DATAGRAM);

          /* Report name-value */
          createIntElement(mqmdChild, (CciChar*)constReport, (CciInt)0);

          /* Create the User Name - Broker's user id name-value */
          _getAttribute(context, (CciChar*)constUserIdString, buffer, sizeof(buffer));
          createCharElement(mqmdChild, (CciChar*)constUserIdentifier, buffer, CciCharLen(buffer));

          /* Create a non transactional message, ie the publication will automatically be put to the Queue */
          createBoolElement(mqmdChild, (CciChar*)constTransactional, (CciBool)0);

          /* Actions under RFH2 parser */

          /* Create the psc folder */
          pscChild = cniCreateElementAsLastChild(&rc, rfh2Child);
          CCI_CHECK_RC();

          cniSetElementName(&rc, pscChild, (CciChar*)constPscFolder);          
          CCI_CHECK_RC();

          /* Create the command name Command name*/
          commandChild = cniCreateElementAsLastChild(&rc, pscChild);
          CCI_CHECK_RC();

          cniSetElementName(&rc, commandChild, (CciChar*)constCommandString);
          CCI_CHECK_RC();

          /* Create the command value */
          commandValueChild = cniCreateElementAsLastChild(&rc, commandChild);
          CCI_CHECK_RC();
                    
          /* Set the character value */
          cniSetElementCharacterValue(&rc,
                                      commandValueChild,
                                      (CciChar*)constPublishString,
                                      constPublishStringLength);
          CCI_CHECK_RC();

          /* Create the topic name - value pair */  
          sizeOfTopicString = constSysTopicLength +
                              broker_info_st.brokerName.dataLength +
                              constTopicExceptionStringLength + 
                              broker_info_st.executionGroupName.dataLength + 
                              constTopicSeperatorLength + 
                              broker_info_st.messageFlowName.dataLength;
                               
          topic = (CciChar *)malloc(sizeOfTopicString * sizeof(CciChar) + 1);

          CciCharCpy(topic, (CciChar*)constSysTopic);
          CciCharCat(topic, broker_info_st.brokerName.buffer);
          CciCharCat(topic, (CciChar*)constTopicExceptionString); 
          CciCharCat(topic, broker_info_st.executionGroupName.buffer);
          CciCharCat(topic, (CciChar*)constTopicSeperator);
          CciCharCat(topic, broker_info_st.messageFlowName.buffer);
          
          topicChild = cniCreateElementAsLastChild(&rc, pscChild);
          CCI_CHECK_RC();

          cniSetElementName(&rc, topicChild, (CciChar*)constTopicString);
          CCI_CHECK_RC();

          /* Create the command value */
          topicValueChild = cniCreateElementAsLastChild(&rc, topicChild);
          CCI_CHECK_RC();

          /* Set the character value */
          cniSetElementCharacterValue(&rc,
                                      topicValueChild,
                                      topic,
                                      CciCharLen(topic));
          CCI_CHECK_RC();
                                       
          /* Actions under the XML parser */
          /* Create broker tag */
          brokerChild = cniCreateElementAsLastChild(&rc, xmlChild);
          CCI_CHECK_RC();

          cniSetElementName(&rc, brokerChild, (CciChar*)constBrokerString);          
          CCI_CHECK_RC();

          /* Create label name value pair */
          createCharElement(brokerChild, (CciChar*)constLabelString, broker_info_st.brokerName.buffer, broker_info_st.brokerName.dataLength - 1);

          /* Create Execution group tag */
          executionGroupChild = cniCreateElementAsLastChild(&rc, brokerChild);
          CCI_CHECK_RC();

          cniSetElementName(&rc, executionGroupChild, (CciChar*)constExecutionGroupString);
          CCI_CHECK_RC();
          
          /* Create label name value pair */
          createCharElement(executionGroupChild, (CciChar*)constLabelString, broker_info_st.executionGroupName.buffer, broker_info_st.executionGroupName.dataLength - 1);

          /* Messageflow Element */
          messageflowChild = cniCreateElementAsLastChild(&rc, executionGroupChild);
          CCI_CHECK_RC();

          cniSetElementName(&rc, messageflowChild, (CciChar*)constMessageFlowString);
          CCI_CHECK_RC();

          /* Create label name value pair */
          createCharElement(messageflowChild, (CciChar*)constLabelString, broker_info_st.messageFlowName.buffer, broker_info_st.messageFlowName.dataLength - 1);

          /* Create Exception tag */
          exceptionChild = cniCreateElementAsLastChild(&rc, messageflowChild);
          CCI_CHECK_RC();

          cniSetElementName(&rc, exceptionChild, (CciChar*)constExceptionString);
          CCI_CHECK_RC();

          /* Fill in exception details under the exception tag */

          /* Create version tag */
          createIntElement(exceptionChild, (CciChar*)constVersionString, exception_st.versionId); 

          /* Create Type */
          createIntElement(exceptionChild, (CciChar*)constTypeString, exception_st.type);

          /* Create Message number */
          createIntElement(exceptionChild, (CciChar*)constMessageNumberString, exception_st.messageNumber);

          /* Inserts */
          for (numInserts = 0; numInserts < exception_st.insertCount; numInserts++)
            {
            createCharElement(exceptionChild, (CciChar*)constInsertString, exception_st.inserts[numInserts].buffer,exception_st.inserts[numInserts].dataLength - 1);
            }

          /* File name */
          fileName = CciString(exception_st.fileName);
          createCharElement(exceptionChild, (CciChar*)constFileNameString, fileName, CciCharLen(fileName));

          /* Line number */
          createIntElement(exceptionChild, (CciChar*)constLineNumberString, exception_st.lineNumber);

          /* Function name */
          exceptionFunctionName = CciString(exception_st.functionName);
          createCharElement(exceptionChild, (CciChar*)constFunctionNameString, exceptionFunctionName, CciCharLen(exceptionFunctionName));

          /* Trace text */
          traceText = CciString(exception_st.traceText);
          createCharElement(exceptionChild, (CciChar*)constTraceTextString, traceText, CciCharLen(traceText));

          /* Object name */
          createCharElement(exceptionChild, (CciChar*)constObjectNameString, exception_st.objectName.buffer, exception_st.objectName.dataLength);

          /* Object type */
          createCharElement(exceptionChild, (CciChar*)constObjectTypeString, exception_st.objectType.buffer, exception_st.objectType.dataLength);
          
          /* Propagate the new message to the publish terminal */
          terminalObject = getOutputTerminalHandle(context, (CciChar*)constPublishTerminal);
          
          /* If the terminal exists and is attached, propagate to it */
          if (terminalObject) 
          {
            if (cniIsTerminalAttached(&rc, terminalObject)) 
            {
              /* As this is a new, and changed message, it should be finalized... */
              cniFinalize(&rc, publishMsg, 0);
              cniWriteBuffer(&rc, publishMsg);
              retvaluePublish = cniPropagate(&rc, terminalObject, destinationList, exceptionList, publishMsg);
            }
          }
         
          /* Release storage */
          if (broker_info_st.brokerName.buffer)         free(broker_info_st.brokerName.buffer);
          if (broker_info_st.executionGroupName.buffer) free(broker_info_st.executionGroupName.buffer);
          if (broker_info_st.messageFlowName.buffer)    free(broker_info_st.messageFlowName.buffer);
          if (broker_info_st.queueManagerName.buffer)   free(broker_info_st.queueManagerName.buffer);
          if (topic) free(topic);
          if (exceptionFunctionName) free(exceptionFunctionName);
          if (fileName) free(fileName);
          if (traceText) free(traceText);
            
          /* Rethrow the last exception. */
          cciRethrowLastException(&rc);
        }
        else 
        {
          /* Any further local error handling can go here */
        }
      }
      else {
        printf("evaluate() message propagated\n");
        fflush(stdout);
      }
    }
  }
  
  /* Delete the message we created now we have finished with it */
  if (publishMsg)
    cniDeleteMessage(0, publishMsg);

  return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */ 
/*  Node definition - ComIbmPublishExceptionNode                              */
/*                                                                            */
/******************************************************************************/
void definePublishExceptionNode(
  void* factoryObject
){
  static char* functionName = "definePublishExceptionNode()";
  static CNI_VFT vftable;

  /* Setup function table with pointers to node implementation functions */
  memset(&vftable, 0, sizeof(vftable));
  vftable.iFpCreateNodeContext = _PublishException_createNodeContext;
  vftable.iFpDeleteNodeContext = _deleteNodeContext;
  vftable.iFpGetAttributeName  = _getAttributeName;
  vftable.iFpSetAttribute      = _setAttribute;
  vftable.iFpGetAttribute      = _getAttribute;
  vftable.iFpEvaluate          = _PublishException_evaluate;

  /* Define a node type supported by our factory. If any errors/exceptions    */
  /* occur during the execution of this utility function, then as we have not */
  /* supplied the returnCode argument, the exception will bypass the plugin   */
  /* and be directly handled by the broker.                                   */

  cniDefineNodeClass(0, factoryObject, (CciChar*)constPublishExceptionNode, &vftable);

  return;
}

/******************************************************************************/
/*                                                                            */
/*  initNodeConstants                                                         */
/*                                                                            */
/*  Creates Cci constant strings from defined char * constant strings.        */
/*                                                                            */
/******************************************************************************/
void initNodeConstants( )
{
  
/* Parser constant strings */
  
  constMqPropertyParser       = CciString(CONST_MQ_PROPERTY_PARSER);
  constMQHMDParser            = CciString(CONST_MQMD_PARSER);
  constMQHRF2Parser           = CciString(CONST_MQHRF2_PARSER);
  constXMLParser              = CciString(CONST_XML_PARSER);

/* MQMD Defined constant strings */

  constUserIdentifier         = CciString(CONST_USER_IDENTIFIER);
  constFormat                 = CciString(CONST_FORMAT);
  constMsgType                = CciString(CONST_MESSAGE_TYPE);
  constReport                 = CciString(CONST_REPORT);
  constTransactional          = CciString(CONST_TRANSACTIONAL);
  constFormatLength           = CciCharLen((CciChar *)constFormat);

/* RFH2 Defined constant strings */

  constPscFolder              = CciString(CONST_SYS_PSC_FOLDER);

/* PubSub Constants */
  
  constSysTopic               = CciString(CONST_SYS_TOPIC);
  constTopicExceptionString   = CciString(CONST_TOPIC_EXCEPTION_STRING);
  constTopicSeperator         = CciString(CONST_TOPIC_SEPERATOR);
  constCommandString          = CciString(CONST_COMMAND_STRING);
  constPublishString          = CciString(CONST_PUBLISH_STRING);
  constTopicString            = CciString(CONST_TOPIC_STRING);
  constPublishStringLength    = CciCharLen((CciChar *)constPublishString);
  constSysTopicLength         = CciCharLen((CciChar *)constSysTopic);
  constTopicExceptionStringLength = CciCharLen((CciChar *)constTopicExceptionString);
  constTopicSeperatorLength   = CciCharLen((CciChar *)constTopicSeperator);

/* XML Constants */

  constBrokerString           = CciString(CONST_BROKER_STRING);
  constLabelString            = CciString(CONST_LABEL_STRING);
  constExecutionGroupString   = CciString(CONST_EXECUTION_GROUP_STRING);
  constMessageFlowString      = CciString(CONST_MESSAGEFLOW_STRING);
  constExceptionString        = CciString(CONST_EXCEPTION_STRING);
  constVersionString          = CciString(CONST_VERSION_STRING);
  constTypeString             = CciString(CONST_TYPE_STRING);
  constMessageNumberString    = CciString(CONST_MESSAGE_NUMBER_STRING);
  constInsertString           = CciString(CONST_INSERT_STRING);
  constFileNameString         = CciString(CONST_FILE_NAME_STRING);
  constLineNumberString       = CciString(CONST_LINE_NUMBER_STRING);
  constFunctionNameString     = CciString(CONST_FUNCTION_NAME_STRING);
  constTraceTextString        = CciString(CONST_TRACE_TEXT_STRING);
  constObjectNameString       = CciString(CONST_OBJECT_NAME_STRING);
  constObjectTypeString       = CciString(CONST_OBJECT_TYPE_STRING);
  
/* Node Constants */

  constPublishExceptionNode   = CciString(CONST_PUBLISH_EXCEPTION_NODE);
  constPublishExceptionNodeFactory = CciString(CONST_PUBLISH_EXCEPTION_NODE_FACTORY);
  constUserIdString           = CciString(CONST_USERID_STRING);
  constZero                   = CciString(CONST_ZERO);

/* Terminal Constants */

  constInTerminal = CciString(CONST_IN_TERMINAL);
  constOutTerminal = CciString(CONST_OUT_TERMINAL);
  constPublishTerminal = CciString(CONST_PUBLISH_TERMINAL);
  constFailureTerminal = CciString(CONST_FAILURE_TERMINAL);

}

/******************************************************************************/
/*                                                                            */
/* Plugin Node Implementation Function:                                       */
/*                                                                            */
/******************************************************************************/
CciFactory LilFactoryExportPrefix * LilFactoryExportSuffix bipGetMessageflowNodeFactory()
{
  static char* functionName = "bipGetMessageflowNodeFactory()";
  CciFactory*      factoryObject;
  int              rc = 0;
  CCI_EXCEPTION_ST exception_st;

  /* Before we proceed we need to initialise all the static constants */
  /* that may be used by the plug-in.                                 */
  initNodeConstants();
  
  /* Create the Node Factory for this plug-in */
  factoryObject = cniCreateNodeFactory(0, (unsigned short *)constPublishExceptionNodeFactory);
  
  if (factoryObject == CCI_NULL_ADDR) {
    if (rc == CCI_EXCEPTION) {
      /* Get details of the exception */
      memset(&exception_st, 0, sizeof(exception_st));
      cciGetLastExceptionData(&rc, &exception_st);
      
      /* Rethrow the exception */
      cciRethrowLastException(&rc);
    }
  }
  else {
    /* Define the nodes supported by this factory */
    definePublishExceptionNode(factoryObject);
  }

  /* Return address of this factory object to the broker */
  return(factoryObject);
}

